from enum import IntEnum

# Kummenberg Internal Definitions
from phootonics_controller.utils.MR_E_2_Python_SDK.optoKummenberg.tools.definitions import *

# MRE2 Specific Definitions
MRE2_CHANNEL_CNT = 2
